from __future__ import annotations
import json
from typing import TYPE_CHECKING

from django.core.serializers.json import DjangoJSONEncoder

from django_glue import constants
from django_glue.conf import settings
from django_glue.session import Session

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest

def django_glue(request: WSGIRequest) -> dict:
    return {
        constants.VERSION_CONTEXT_NAME: constants.__VERSION__,
        constants.KEEP_LIVE_INTERVAL_TIME_MILLISECONDS_CONTEXT_NAME: (settings.DJANGO_GLUE_KEEP_LIVE_EXPIRE_TIME_SECONDS / 2.2) * 1000,
        constants.SESSION_DATA_CONTEXT_NAME: json.dumps(Session(request).session, cls=DjangoJSONEncoder),
    }



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from django.utils import timezone

if TYPE_CHECKING:
    from django_glue.glue.model_object.fields.glue import ModelFieldGlue


def serialize_field_value(model_field_glue: ModelFieldGlue) -> Any:
    formatted_value = model_field_glue.value

    if formatted_value is not None:

        if model_field_glue._meta.type == 'DateTimeField':
            try:
                formatted_value = timezone.localtime(model_field_glue.value).strftime('%Y-%m-%dT%H:%M')
            except Exception:
                formatted_value = model_field_glue.value.strftime('%Y-%m-%dT%H:%M')
        elif model_field_glue._meta.type == 'DateField':
            try:
                formatted_value = timezone.localdate(model_field_glue.value).strftime('%Y-%m-%d')
            except Exception:
                formatted_value = model_field_glue.value.strftime('%Y-%m-%d')

    return formatted_value



import json

from django.http import JsonResponse, HttpRequest
from django.views.decorators.http import require_http_methods

from django_glue.decorators import require_content_types
from django_glue.handler.body import RequestBody
from django_glue.handler.utils import process_request
from django_glue.response.responses import generate_json_404_response
from django_glue.session import KeepLiveSession, Session


@require_http_methods(["POST"])
@require_content_types('application/json', 'text/html')
def handler_ajax_view(request: HttpRequest) -> JsonResponse:
    session = Session(request)
    request_body = RequestBody(request.body)

    if request_body.unique_name in session.session:
        return process_request(session, request_body).to_django_json_response()
    else:
        return generate_json_404_response()


def keep_live_handler_ajax_view(request: HttpRequest) -> JsonResponse:
    data = json.loads(request.body)
    unique_names = data.get('unique_names', [])

    if len(unique_names) > 0:
        KeepLiveSession(request).update_unique_names(unique_names)

    return JsonResponse(
        data=Session(request).session
    )


def session_data_ajax_view(request: HttpRequest) -> JsonResponse:
    session = Session(request)
    return JsonResponse(
        data=session.session
    )





from django.urls import path

from django_glue import views_ajax, constants

app_name = constants.URL_APP_NAME

urlpatterns = [
    # These url path names are used in our middleware to avoid cleaning session data.
    path("", views_ajax.handler_ajax_view, name=constants.HANDLER_URL_NAME),
    path("keep_live/", views_ajax.keep_live_handler_ajax_view, name=constants.KEEP_LIVE_HANDLER_URL_NAME),
    path("session/", views_ajax.session_data_ajax_view, name=constants.SESSION_DATA_URL_NAME),
]



from django_glue.access.access import Access
from django_glue.access.actions import BaseAction


class FunctionGlueAction(BaseAction):
    CALL = 'call'

    def required_access(self) -> Access:
        if self.value == 'call':
            return Access.VIEW
        else:
            raise ValueError('That is not a valid action on a glue function.')



from django_glue.access.access import Access
from django_glue.glue.enums import GlueType
from django_glue.glue.function.response_data import FunctionGlueJsonData
from django_glue.glue.function.session_data import FunctionGlueSessionData
from django_glue.glue.glue import BaseGlue
from django_glue.utils import check_valid_method_kwargs, type_set_method_kwargs


class FunctionGlue(BaseGlue):
    def __init__(
            self,
            unique_name: str,
            function_path: str,
    ):
        super().__init__(unique_name, GlueType.FUNCTION, Access.VIEW)
        self.function_path = function_path

        self.module_name = '.'.join(function_path.split('.')[:-1])
        self.function_name = function_path.split('.')[-1]

    def call(self, function_kwargs: dict):
        module = __import__(self.module_name, fromlist=[self.function_name])

        if hasattr(module, self.function_name):
            function = getattr(module, self.function_name)

            if check_valid_method_kwargs(function, function_kwargs):
                type_set_kwargs = type_set_method_kwargs(function, function_kwargs)

                return function(**type_set_kwargs)

        return None

    def to_session_data(self) -> FunctionGlueSessionData:
        return FunctionGlueSessionData(
            unique_name=self.unique_name,
            glue_type=self.glue_type,
            access=self.access,
            function_path=self.function_path
        )

    @staticmethod
    def to_response_data(function_return) -> FunctionGlueJsonData:
        return FunctionGlueJsonData(function_return)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django_glue.access.access import Access


class BaseAction(str, Enum):

    def __str__(self) -> str:
        return self.value

    def required_access(self) -> Access:
        raise NotImplemented('You must override required access on Glue Actions')


from django_glue.access.decorators import check_access
from django_glue.glue.model_object.response_data import ModelObjectGlueJsonData, MethodModelObjectGlueJsonData
from django_glue.glue.model_object.tools import model_object_glue_from_query_set_glue_session_data, \
    model_object_glues_from_query_set_glue_and_session_data
from django_glue.glue.post_data import GetPostData, DeletePostData, UpdatePostData, MethodPostData
from django_glue.glue.query_set.actions import QuerySetGlueAction
from django_glue.glue.query_set.post_data import FilterQuerySetGluePostData
from django_glue.glue.query_set.response_data import QuerySetGlueJsonData, MethodQuerySetGlueJsonData, \
    ToChoicesQuerySetGlueJsonData
from django_glue.glue.query_set.session_data import QuerySetGlueSessionData
from django_glue.glue.query_set.tools import query_set_glue_from_session_data
from django_glue.handler.handlers import BaseRequestHandler
from django_glue.response.data import JsonResponseData
from django_glue.response.responses import generate_json_200_response_data


class AllQuerySetGlueHandler(BaseRequestHandler):
    action = QuerySetGlueAction.ALL
    _session_data_class = QuerySetGlueSessionData

    @check_access
    def process_response_data(self) -> JsonResponseData:

        query_set_glue = query_set_glue_from_session_data(self.session_data)
        glue_model_objects = model_object_glues_from_query_set_glue_and_session_data(query_set_glue.query_set.all(), self.session_data)

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=QuerySetGlueJsonData([ModelObjectGlueJsonData(glue_model_object.fields) for glue_model_object in glue_model_objects])
        )


class DeleteGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.DELETE
    _session_data_class = QuerySetGlueSessionData
    _post_data_class = DeletePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        query_set_glue = query_set_glue_from_session_data(self.session_data)

        filtered_query_set = query_set_glue.query_set.filter(id__in=self.post_data.id)
        filtered_query_set.delete()

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully deleted queryset!',
        )


class FilterGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.FILTER
    _session_data_class = QuerySetGlueSessionData
    _post_data_class = FilterQuerySetGluePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:

        query_set_glue = query_set_glue_from_session_data(self.session_data)
        filtered_query_set = query_set_glue.query_set.filter(**self.post_data.filter_params)
        model_objects_glue = model_object_glues_from_query_set_glue_and_session_data(filtered_query_set, self.session_data)

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=QuerySetGlueJsonData([ModelObjectGlueJsonData(glue_model_object.fields) for glue_model_object in model_objects_glue])
        )


class GetGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.GET
    _session_data_class = QuerySetGlueSessionData
    _post_data_class = GetPostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        query_set_glue = query_set_glue_from_session_data(self.session_data)

        model_object = query_set_glue.query_set.get(id=self.post_data.id)
        model_object_glue = model_object_glue_from_query_set_glue_session_data(model_object, self.session_data)

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=QuerySetGlueJsonData([ModelObjectGlueJsonData(model_object_glue.fields)])
        )


class NullObjectGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.NULL_OBJECT
    _session_data_class = QuerySetGlueSessionData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        query_set_glue = query_set_glue_from_session_data(self.session_data)
        model_object_glue = model_object_glue_from_query_set_glue_session_data(query_set_glue.model(), self.session_data)

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=QuerySetGlueJsonData([ModelObjectGlueJsonData(model_object_glue.fields)])
        )


class MethodGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.METHOD
    _session_data_class = QuerySetGlueSessionData
    _post_data_class = MethodPostData

    def process_response_data(self) -> JsonResponseData:
        query_set_glue = query_set_glue_from_session_data(self.session_data)

        if isinstance(self.post_data.id, int):
            # Called from a glue queryset on a model object
            model_object = query_set_glue.query_set.get(id=self.post_data.id)
            model_object_glue = model_object_glue_from_query_set_glue_session_data(model_object, self.session_data)
            method_return = model_object_glue.call_method(self.post_data.method, self.post_data.kwargs)

            return generate_json_200_response_data(
                message_title='Success',
                message_body='Successfully updated model object!',
                data=MethodModelObjectGlueJsonData(method_return)
            )
        else:
            # Called from a glue queryset
            filtered_query_set = query_set_glue.query_set.filter(id__in=self.post_data.id)

            method_return_data = []

            for model_object in filtered_query_set:
                model_object_glue = model_object_glue_from_query_set_glue_session_data(model_object, self.session_data)
                method_return = model_object_glue.call_method(self.post_data.method, self.post_data.kwargs)
                method_return_data.append(MethodModelObjectGlueJsonData(method_return))

            return generate_json_200_response_data(
                message_title='Success',
                message_body='Successfully updated model object!',
                data=MethodQuerySetGlueJsonData(method_return_data)
            )


class UpdateGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.UPDATE
    _session_data_class = QuerySetGlueSessionData
    _post_data_class = UpdatePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        query_set_glue = query_set_glue_from_session_data(self.session_data)

        model_object = query_set_glue.query_set.get(id=self.post_data.id)

        model_object_glue = model_object_glue_from_query_set_glue_session_data(model_object, self.session_data)
        model_object_glue.update(self.post_data.fields)

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully updated model object!',
            data=QuerySetGlueJsonData([ModelObjectGlueJsonData(model_object_glue.fields)])
        )


class ToChoicesGlueQuerySetHandler(BaseRequestHandler):
    action = QuerySetGlueAction.TO_CHOICES
    _session_data_class = QuerySetGlueSessionData
    _post_data_class = FilterQuerySetGluePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        query_set_glue = query_set_glue_from_session_data(self.session_data)
        filtered_query_set = query_set_glue.query_set.filter(**self.post_data.filter_params)
        model_object_glues = model_object_glues_from_query_set_glue_and_session_data(filtered_query_set, self.session_data)
        choices = [(model_object_glue.model_object.pk, str(model_object_glue.model_object)) for model_object_glue in model_object_glues]

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=ToChoicesQuerySetGlueJsonData(choices)
        )



import json
from typing import Any, Iterable

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpRequest

from django_glue.conf import settings
from django_glue.glue.glue import BaseGlue
from django_glue.session.data import SessionData


class Session:
    """
        Used to add models, query sets, and other objects to the session.
    """
    def __init__(self, request: HttpRequest):
        self.request = request
        self.request.session.setdefault(settings.DJANGO_GLUE_SESSION_NAME, dict())
        self.session = self.request.session[settings.DJANGO_GLUE_SESSION_NAME]

    def __getitem__(self, key: str) -> Any:
        return self.session[key]

    def __setitem__(self, key: str, value: Any):
        self.session[key] = value

    def add_glue(self, glue: BaseGlue) -> None:
        if glue.unique_name in self.session:
            self.purge_unique_name(glue.unique_name)

        self.add_session_data(glue.unique_name, glue.to_session_data())
        self.set_modified()

    def add_session_data(self, unique_name: str, session_data: SessionData) -> None:
        self.session[unique_name] = session_data.to_dict()

    def clean(self, removable_unique_names: Iterable[str]) -> None:
        for unique_name in removable_unique_names:
            self.purge_unique_name(unique_name)

        self.set_modified()

    def purge_unique_name(self, unique_name: str) -> None:
        self.session.pop(unique_name)

    def set_modified(self) -> None:
        self.request.session.modified = True

    def to_json(self) -> str:
        return json.dumps(self.session, cls=DjangoJSONEncoder)



from dataclasses import dataclass
from typing import Any

from django_glue.response.data import BaseJsonData


@dataclass
class FunctionGlueJsonData(BaseJsonData):
    function_return: Any

    def to_dict(self) -> dict:
        return {'function_return': self.function_return}



from django_glue.glue.enums import GlueType
from django_glue.glue.function.maps import FUNCTION_GLUE_HANDLER_MAP
from django_glue.glue.model_object.maps import MODEL_OBJECT_GLUE_HANDLER_MAP
from django_glue.glue.query_set.maps import QUERY_SET_GLUE_HANDLER_MAP
from django_glue.glue.template.maps import TEMPLATE_GLUE_HANDLER_MAP


GLUE_TYPE_TO_HANDLER_MAP = {
    GlueType.MODEL_OBJECT: MODEL_OBJECT_GLUE_HANDLER_MAP,
    GlueType.QUERY_SET: QUERY_SET_GLUE_HANDLER_MAP,
    GlueType.FUNCTION: FUNCTION_GLUE_HANDLER_MAP,
    GlueType.TEMPLATE: TEMPLATE_GLUE_HANDLER_MAP,
}



from dataclasses import dataclass
from typing import Union

from django_glue.session.data import SessionData


@dataclass
class QuerySetGlueSessionData(SessionData):
    query_set_str: str
    fields: dict
    included_fields: Union[list, tuple]
    excluded_fields: Union[list, tuple]
    included_methods: Union[list, tuple]



from django_glue.glue.query_set import handlers
from django_glue.glue.query_set.actions import QuerySetGlueAction

QUERY_SET_GLUE_HANDLER_MAP = {
    QuerySetGlueAction.ALL: handlers.AllQuerySetGlueHandler,
    QuerySetGlueAction.DELETE: handlers.DeleteGlueQuerySetHandler,
    QuerySetGlueAction.FILTER: handlers.FilterGlueQuerySetHandler,
    QuerySetGlueAction.GET: handlers.GetGlueQuerySetHandler,
    QuerySetGlueAction.NULL_OBJECT: handlers.NullObjectGlueQuerySetHandler,
    QuerySetGlueAction.METHOD: handlers.MethodGlueQuerySetHandler,
    QuerySetGlueAction.UPDATE: handlers.UpdateGlueQuerySetHandler,
    QuerySetGlueAction.TO_CHOICES: handlers.ToChoicesGlueQuerySetHandler,
}



from django_glue.glue.template import handlers
from django_glue.glue.template.actions import TemplateGlueAction

TEMPLATE_GLUE_HANDLER_MAP = {
    TemplateGlueAction.GET: handlers.GetTemplateGlueHandler,
}



from typing import Type

from django.apps import apps
from django.db.models import Model, QuerySet

from django_glue.glue.model_object.glue import ModelObjectGlue
from django_glue.glue.model_object.session_data import ModelObjectGlueSessionData
from django_glue.glue.query_set.session_data import QuerySetGlueSessionData


def model_object_glue_from_session_data(session_data: ModelObjectGlueSessionData) -> ModelObjectGlue:
    model: type | Type[Model] = apps.get_model(session_data.app_label, session_data.model_name)

    try:
        model_object = model.objects.get(pk=session_data.object_pk)
    except model.DoesNotExist:
        model_object = model()

    return ModelObjectGlue(
        unique_name=session_data.unique_name,
        model_object=model_object,
        access=session_data.access,
        included_fields=session_data.included_fields,
        excluded_fields=session_data.exclude_fields,
        included_methods=session_data.methods,
    )


def model_object_glue_from_query_set_glue_session_data(
        model_object: Model,
        session_data: QuerySetGlueSessionData
):
    return ModelObjectGlue(
        unique_name=session_data.unique_name,
        access=session_data.access,
        model_object=model_object,
        included_fields=session_data.included_fields,
        excluded_fields=session_data.excluded_fields,
        included_methods=session_data.included_methods,
    )


def model_object_glues_from_query_set_glue_and_session_data(
        query_set: QuerySet,
        session_data: QuerySetGlueSessionData
) -> list[ModelObjectGlue]:

    return [model_object_glue_from_query_set_glue_session_data(
        model_object,
        session_data
    ) for model_object in query_set]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from django_glue.access.access import Access
from django_glue.glue.post_data import BasePostData

if TYPE_CHECKING:
    from django_glue.session import Session
    from django_glue.handler.body import RequestBody
    from django_glue.response.data import JsonResponseData
    from django_glue.access.actions import BaseAction
    from django_glue.session.data import SessionData


class BaseRequestHandler(ABC):
    action: BaseAction | None = None
    _session_data_class: SessionData | None = None
    _post_data_class: BasePostData | None = None

    def __init__(self, session: Session, request_body: RequestBody):
        if self._session_data_class is None:
            raise ValueError(f'Please initialize class variable _session_data_class on {self.__class__.__name__}')

        if self.action is None:
            raise ValueError(f'Please initialize class variable action on {self.__class__.__name__}')

        self.unique_name = request_body.unique_name  # Unique name maps what the user is requesting

        if self._post_data_class is None:
            self.post_data = request_body.data
        else:
            self.post_data = self._post_data_class(**request_body.data['data'])  # The data we are expecting in post

        self.session_data = self._session_data_class(**session[self.unique_name])  # data we stored in glue session.

    def has_access(self) -> bool:
        access = Access(self.session_data.access)
        return access.has_access(self.action.required_access())

    @abstractmethod
    def process_response_data(self) -> JsonResponseData:
        # Todo: Do we want to handle an error message here or let the system crash?
        pass



from typing import Type

from django.db.models import Model

from django_glue.form.field.factories import FormFieldFactory
from django_glue.glue.model_object.fields.glue import ModelFieldsGlue, ModelFieldGlue, ModelFieldMetaGlue
from django_glue.glue.model_object.fields.utils import field_name_included


def model_object_fields_glue_from_model(
        model: Type[Model],
        included_fields: tuple,
        excluded_fields: tuple
) -> ModelFieldsGlue:
    fields = []

    for model_field in model._meta.fields:
        if field_name_included(model_field.name, included_fields, excluded_fields):

            _meta = ModelFieldMetaGlue(
                type=model_field.get_internal_type(),
                name=model_field.name,
                glue_field=FormFieldFactory(model_field).factory_method()
            )

            glue_model_field = ModelFieldGlue(
                name=model_field.name,
                value=None,
                _meta=_meta
            )

            fields.append(glue_model_field)

    return ModelFieldsGlue(fields=fields)



from django_glue.shortcuts import glue_model_object, glue_function, glue_query_set, glue_template


__all__ = [
    'glue_model_object',
    'glue_function',
    'glue_query_set',
    'glue_template'
]


from dataclasses import dataclass
from typing import Any

from django_glue.glue.model_object.fields.glue import ModelFieldsGlue
from django_glue.response.data import BaseJsonData


@dataclass
class ModelObjectGlueJsonData(BaseJsonData):  # This is a little duplicated but allows us to send more response data.
    fields: ModelFieldsGlue

    def to_dict(self) -> dict:
        return self.fields.to_dict()


@dataclass
class MethodModelObjectGlueJsonData(BaseJsonData):
    method_return: Any

    def to_dict(self) -> dict:
        return {'method_return': self.method_return}



from django_glue.form.field.attributes import factories
from django_glue.form.field.enums import FieldType


FIELD_TYPE_TO_FIELD_ATTRIBUTE_FACTORY_MAP = {
    FieldType.AUTO_FIELD: factories.IntegerAttributeFactory,
    FieldType.BIG_AUTO_FIELD: factories.IntegerAttributeFactory,
    FieldType.SMALL_AUTO_FIELD: factories.IntegerAttributeFactory,
    FieldType.BOOLEAN: factories.BooleanAttributeFactory,
    FieldType.CHAR: factories.CharAttributeFactory,
    FieldType.COMMA_SEPARATED_INTEGER: factories.IntegerAttributeFactory,  # Now depreciated in django.
    FieldType.DATE: factories.DateAttributeFactory,
    FieldType.DATETIME: factories.DateAttributeFactory,
    FieldType.DECIMAL: factories.DecimalAttributeFactory,
    FieldType.DURATION: factories.CharAttributeFactory,
    FieldType.EMAIL: factories.CharAttributeFactory,
    FieldType.FLOAT: factories.DecimalAttributeFactory,
    FieldType.INTEGER: factories.IntegerAttributeFactory,
    FieldType.BIG_INTEGER: factories.IntegerAttributeFactory,
    FieldType.SMALL_INTEGER: factories.IntegerAttributeFactory,
    FieldType.GENERIC_IP_ADDRESS: factories.CharAttributeFactory,
    FieldType.IP_ADDRESS: factories.CharAttributeFactory,
    FieldType.FOREIGN_KEY: factories.IntegerAttributeFactory,
    FieldType.ONE_TO_ONE: factories.IntegerAttributeFactory,
    FieldType.MANY_TO_MANY: factories.CharAttributeFactory,
    FieldType.JSON: factories.TextAreaAttributeFactory,
    FieldType.NULL_BOOLEAN: factories.BooleanAttributeFactory,  # Now depreciated in django
    FieldType.POSITIVE_BIG_INTEGER: factories.IntegerAttributeFactory,
    FieldType.POSITIVE_INTEGER: factories.IntegerAttributeFactory,
    FieldType.POSITIVE_SMALL_INTEGER: factories.IntegerAttributeFactory,
    FieldType.SLUG: factories.CharAttributeFactory,
    FieldType.TEXT: factories.TextAreaAttributeFactory,
    FieldType.TIME: factories.DateAttributeFactory,
    FieldType.URL: factories.CharAttributeFactory,
    FieldType.UUID: factories.CharAttributeFactory
}



from django_glue.access.decorators import check_access
from django_glue.glue.function.actions import FunctionGlueAction
from django_glue.glue.function.glue import FunctionGlue
from django_glue.glue.function.post_data import CallFunctionGluePostData

from django_glue.glue.function.session_data import FunctionGlueSessionData
from django_glue.handler.handlers import BaseRequestHandler
from django_glue.response.data import JsonResponseData
from django_glue.response.responses import generate_json_200_response_data


class CallFunctionGlueHandler(BaseRequestHandler):
    action = FunctionGlueAction.CALL
    _session_data_class = FunctionGlueSessionData
    _post_data_class = CallFunctionGluePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        function_glue = FunctionGlue(
            unique_name=self.session_data.unique_name,
            function_path=self.session_data.function_path
        )

        function_return = function_glue.call(self.post_data.kwargs)

        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=function_glue.to_response_data(function_return)
        )




from django_glue.glue.model_object import handlers
from django_glue.glue.model_object.actions import ModelObjectGlueAction

MODEL_OBJECT_GLUE_HANDLER_MAP = {
    ModelObjectGlueAction.GET: handlers.GetModelObjectGlueHandler,
    ModelObjectGlueAction.UPDATE: handlers.UpdateModelObjectGlueHandler,
    ModelObjectGlueAction.METHOD: handlers.MethodModelObjectGlueHandler,
    ModelObjectGlueAction.DELETE: handlers.DeleteModelObjectGlueHandler
}



from django_glue.constants import GLUE_TYPE_KEY
from django_glue.glue.enums import GlueType
from django_glue.handler.body import RequestBody
from django_glue.handler.maps import GLUE_TYPE_TO_HANDLER_MAP
from django_glue.response.data import JsonResponseData
from django_glue.session import Session


def process_request(session: Session, request_body: RequestBody) -> JsonResponseData:
    # Todo: Validation errors.
    glue_type = GlueType(session[request_body.unique_name][GLUE_TYPE_KEY])

    handler_class = GLUE_TYPE_TO_HANDLER_MAP[GlueType(glue_type)][request_body.action]

    return handler_class(
        session=session,
        request_body=request_body
    ).process_response_data()



from django_glue.access.decorators import check_access
from django_glue.glue.template.actions import TemplateGlueAction
from django_glue.glue.template.glue import TemplateGlue
from django_glue.glue.template.post_data import GetTemplateGluePostData
from django_glue.glue.template.session_data import TemplateSessionData
from django_glue.handler.handlers import BaseRequestHandler
from django_glue.response.data import JsonResponseData
from django_glue.response.responses import generate_json_200_response_data


class GetTemplateGlueHandler(BaseRequestHandler):
    action = TemplateGlueAction.GET
    _session_data_class = TemplateSessionData
    _post_data_class = GetTemplateGluePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        template_glue = TemplateGlue(
            unique_name=self.session_data.unique_name,
            template_name=self.session_data.template_name
        )
        rendered_template = template_glue.render_to_string(self.post_data.context_data)
        return generate_json_200_response_data(
            'THE METHOD ACTION',
            'this is a response from an model object method action!',
            data=template_glue.to_response_data(rendered_template).to_dict()
        )



from typing import Union

from django.db.models import Model
from django.db.models.query import QuerySet
from django.http import HttpRequest

from django_glue.constants import ALL_DUNDER_KEY, NONE_DUNDER_KEY
from django_glue.glue.function.glue import FunctionGlue
from django_glue.glue.glue import BaseGlue
from django_glue.glue.model_object.glue import ModelObjectGlue
from django_glue.glue.query_set.glue import QuerySetGlue
from django_glue.glue.template.glue import TemplateGlue
from django_glue.session import Session, KeepLiveSession
from django_glue.utils import encode_unique_name


def _glue_base_function(request: HttpRequest, glue: BaseGlue) -> None:

    glue_session = Session(request)
    glue_session.add_glue(glue)

    glue_keep_live_session = KeepLiveSession(request)
    glue_keep_live_session.set_unique_name(glue.unique_name)

    # Todo: Check to see if this has actually changed.
    glue_session.set_modified()


def glue_function(
        request: HttpRequest,
        unique_name: str,
        target: str,
) -> None:
    glue_function_entity = FunctionGlue(
        unique_name=encode_unique_name(request, unique_name),
        function_path=target
    )
    _glue_base_function(request, glue_function_entity)


def glue_model_object(
        request: HttpRequest,
        unique_name: str,
        model_object: Model,
        access: str = 'view',
        fields: Union[list, tuple] = (ALL_DUNDER_KEY,),
        exclude: Union[list, tuple] = (NONE_DUNDER_KEY,),
        methods: Union[list, tuple] = (NONE_DUNDER_KEY,),
) -> None:
    glue_model_object_entity = ModelObjectGlue(
        unique_name=encode_unique_name(request, unique_name),
        model_object=model_object,
        access=access,
        included_fields=fields,
        excluded_fields=exclude,
        included_methods=methods
    )

    _glue_base_function(request, glue_model_object_entity)


def glue_query_set(
        request: HttpRequest,
        unique_name: str,
        target: QuerySet,
        access: str = 'view',
        fields: Union[list, tuple] = (ALL_DUNDER_KEY,),
        exclude: Union[list, tuple] = (NONE_DUNDER_KEY,),
        methods: Union[list, tuple] = (NONE_DUNDER_KEY,),
) -> None:
    glue_query_set_entity = QuerySetGlue(
        unique_name=encode_unique_name(request, unique_name),
        query_set=target,
        access=access,
        included_fields=fields,
        excluded_fields=exclude,
        included_methods=methods
    )

    _glue_base_function(request, glue_query_set_entity)


def glue_template(
        request: HttpRequest,
        unique_name: str,
        target: str,
) -> None:

    glue_template_entity = TemplateGlue(
        unique_name=encode_unique_name(request, unique_name),
        template_name=target
    )

    _glue_base_function(request, glue_template_entity)



from time import time
from typing import Any, Iterable

from django.http import HttpRequest

from django_glue.conf import settings


class KeepLiveSession:
    """
        Used to keep glue session data live for a set amount of time.
        Functionality to handle multiple windows/tabs.
    """
    def __init__(self, request: HttpRequest):
        self.request = request
        self.session = request.session.setdefault(settings.DJANGO_GLUE_KEEP_LIVE_SESSION_NAME, dict())

    def __getitem__(self, key: str) -> Any:
        return self.session[key]

    def __setitem__(self, key: str, value: Any):
        self.session[key] = value

    def clean_and_get_expired_unique_names(self) -> list:
        expired_unique_names = []

        for key, val in self.session.items():
            if time() > val:
                expired_unique_names.append(key)

        for expired_unique_name in expired_unique_names:
            self.session.pop(expired_unique_name)

        self.set_modified()

        return expired_unique_names

    @staticmethod
    def get_next_expire_time() -> float:
        return time() + settings.DJANGO_GLUE_KEEP_LIVE_EXPIRE_TIME_SECONDS

    def set_unique_name(self, unique_name: str) -> None:
        self.session.setdefault(unique_name, self.get_next_expire_time())
        self.set_modified()

    def set_modified(self) -> None:
        self.request.session.modified = True

    def update_unique_names(self, unique_names: Iterable[str]) -> None:
        for unique_name in unique_names:
            if unique_name in self.session:
                self.session[unique_name] = self.get_next_expire_time()

        self.set_modified()



from dataclasses import field, dataclass
from typing import Union

from django_glue.glue.model_object.response_data import ModelObjectGlueJsonData, MethodModelObjectGlueJsonData
from django_glue.response.data import BaseJsonData


@dataclass
class QuerySetGlueJsonData(BaseJsonData):
    model_objects: list[ModelObjectGlueJsonData] = field(default_factory=list)

    def to_dict(self) -> list:
        return [model_object.to_dict() for model_object in self.model_objects]


@dataclass
class MethodQuerySetGlueJsonData(BaseJsonData):
    method_returns: list[MethodModelObjectGlueJsonData] = field(default_factory=list)

    def to_dict(self) -> list:
        return [method_return.to_dict() for method_return in self.method_returns]


@dataclass
class ToChoicesQuerySetGlueJsonData(BaseJsonData):
    choices: list[Union[tuple, list]]

    def to_dict(self):
        return self.choices



from django.db.models import Model

from django_glue.constants import ALL_DUNDER_KEY, NONE_DUNDER_KEY
from django_glue.glue.model_object.fields.glue import ModelFieldGlue


def field_name_included(
        name: str,
        fields: list | tuple,
        exclude: list | tuple,
) -> bool:
    included = False

    if name not in exclude or exclude[0] == NONE_DUNDER_KEY:
        if name in fields or fields[0] == ALL_DUNDER_KEY:
            included = True

    return included


def get_field_value_from_model_object(
        model_object: Model,
        model_field_glue: ModelFieldGlue
) -> str:
    relational = ['ForeignKey', 'BinaryField', 'OnetoOneField']
    if model_field_glue._meta.type in relational:
        return getattr(model_object, f'{model_field_glue.name}_id')
    else:
        return getattr(model_object, model_field_glue.name)



from __future__ import annotations

import inspect
import json
import urllib.parse
from typing import Optional, Callable, Union

from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Model
from django.http import HttpRequest


def check_valid_method_kwargs(method: Callable, kwargs: Optional[dict]) -> bool:
    for kwarg in kwargs:
        if kwarg not in inspect.signature(method).parameters.keys():
            return False
    return True


def encode_unique_name(request: HttpRequest, unique_name: str) -> str:
    if 'glue_encode_path' in request.GET:
        encode_path = request.GET['glue_encode_path']
    else:
        encode_path = request.path_info

    return urllib.parse.quote(f'{unique_name}|{encode_path}', safe='')


def serialize_to_json(data: Union[dict, tuple, list, str, int, float, ]) -> str:
    return json.dumps(data, cls=DjangoJSONEncoder)


def type_set_method_kwargs(method: Callable, kwargs: Optional[dict]) -> dict:
    type_set_kwargs = {}

    # This is a dict consisting of all kwargs and there type annotations (If they have type annotations)
    annotations = inspect.getfullargspec(method).annotations

    for kwarg in kwargs:
        if kwarg in annotations:
            # Converts the kwarg to match the type specified in on the methods kwargs
            type_set_kwargs[kwarg] = inspect.getfullargspec(method).annotations[kwarg](kwargs[kwarg])
        else:
            # If there is not a type annotation, the value remains the same
            type_set_kwargs[kwarg] = kwargs[kwarg]

    return type_set_kwargs


def camel_to_snake(string: str) -> str:
    return ''.join(['_' + c.lower() if c.isupper() else c for c in string]).lstrip('_')


def serialize_object_to_json(model_object: Model) -> str:
    return serializers.serialize('json', [model_object, ])



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union, TYPE_CHECKING

from django_glue.access.access import Access
from django_glue.glue.enums import GlueType

if TYPE_CHECKING:
    from django_glue.session.data import SessionData


class BaseGlue(ABC):

    def __init__(
            self,
            unique_name: str,
            glue_type: GlueType,
            access: Union[Access, str] = Access.VIEW,
    ):

        self.unique_name = unique_name
        self.glue_type = glue_type

        if isinstance(access, str):
            self.access = Access(access)
        else:
            self.access = access

    @abstractmethod
    def to_session_data(self) -> SessionData:
        pass



from __future__ import annotations
from django.http import HttpResponse
from django.http import JsonResponse
from typing import TYPE_CHECKING, Callable, Any

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest

def require_content_types(*content_types) -> Callable:
    def decorator(view_func: Callable) -> Callable:
        def _wrapped_view(request: WSGIRequest, *args, **kwargs) -> Any:
            if request.content_type not in content_types:
                return JsonResponse({'error': 'Unsupported Media Type'}, status=415)
            return view_func(request, *args, **kwargs)
        return _wrapped_view
    return decorator


def require_query_method(view_func: Callable) -> Callable:
    def _wrapped_view(request: WSGIRequest, *args, **kwargs) -> Any:
        if request.method != 'QUERY':
            return HttpResponse('Method Not Allowed', status=405)
        return view_func(request, *args, **kwargs)
    return _wrapped_view



__VERSION__ = '0.8.8'

UNIQUE_NAME_KEY = 'unique_name'
ACTION_KEY = 'action'
GLUE_TYPE_KEY = 'glue_type'

# Fields
ALL_DUNDER_KEY = '__all__'
NONE_DUNDER_KEY = '__none__'

# Context
VERSION_CONTEXT_NAME = 'DJANGO_GLUE_VERSION'
KEEP_LIVE_INTERVAL_TIME_MILLISECONDS_CONTEXT_NAME = 'DJANGO_GLUE_KEEP_LIVE_INTERVAL_TIME_MILLISECONDS'
SESSION_DATA_CONTEXT_NAME = 'DJANGO_GLUE_SESSION_DATA'

# Url
URL_APP_NAME = 'django_glue'
HANDLER_URL_NAME = 'handler'
KEEP_LIVE_HANDLER_URL_NAME = 'keep_live_handler'
SESSION_DATA_URL_NAME = 'session_data'



from django.db import models

from django_glue.form.field.attributes.attributes import FieldAttributes
from django_glue.form.field.attributes.maps import FIELD_TYPE_TO_FIELD_ATTRIBUTE_FACTORY_MAP


def field_attr_from_model_field(model_field: models.Field) -> FieldAttributes:
    try:
        factory = FIELD_TYPE_TO_FIELD_ATTRIBUTE_FACTORY_MAP[model_field.__class__.__name__]
    except KeyError:
        factory = FIELD_TYPE_TO_FIELD_ATTRIBUTE_FACTORY_MAP[models.CharField.__name__]

    return factory(model_field).factory_method()



from typing import Union, Any, Callable

from django.db.models import Model

from django_glue.access.access import Access
from django_glue.constants import ALL_DUNDER_KEY, NONE_DUNDER_KEY
from django_glue.glue.enums import GlueType
from django_glue.glue.glue import BaseGlue
from django_glue.glue.model_object.fields.glue import ModelFieldsGlue
from django_glue.glue.model_object.fields.tools import model_object_fields_glue_from_model
from django_glue.glue.model_object.fields.utils import get_field_value_from_model_object
from django_glue.glue.model_object.session_data import ModelObjectGlueSessionData
from django_glue.utils import check_valid_method_kwargs, type_set_method_kwargs


class ModelObjectGlue(BaseGlue):
    def __init__(
            self,
            unique_name: str,
            model_object: Model,
            access: Union[Access, str] = Access.VIEW,
            included_fields: tuple = (ALL_DUNDER_KEY,),
            excluded_fields: tuple = (NONE_DUNDER_KEY,),
            included_methods: tuple = (NONE_DUNDER_KEY,),
    ):
        super().__init__(unique_name, GlueType.MODEL_OBJECT, access)

        self.model_object = model_object
        self.model = model_object._meta.model

        self.included_fields = included_fields
        self.excluded_fields = excluded_fields
        self.included_methods = included_methods

        self.fields: ModelFieldsGlue = self.generate_field_data()

    def call_method(self, method_name: str, method_kwargs: dict) -> Callable | None:
        if method_name in self.included_methods and hasattr(self.model, method_name):
            method = getattr(self.model_object, method_name)

            if check_valid_method_kwargs(method, method_kwargs):
                type_set_kwargs = type_set_method_kwargs(method, method_kwargs)

                return method(**type_set_kwargs)

        return None

    def generate_field_data(self, include_values: bool = True) -> ModelFieldsGlue:

        model_fields = model_object_fields_glue_from_model(
            model=self.model,
            included_fields=self.included_fields,
            excluded_fields=self.excluded_fields
        )

        if include_values:
            for field in model_fields:
                field.value = get_field_value_from_model_object(self.model_object, field)

        return model_fields

    def generate_method_data(self) -> list[str]:
        methods_list = list()

        for method in self.included_methods:
            if hasattr(self.model_object, method):
                methods_list.append(method)
            elif method == NONE_DUNDER_KEY:
                pass
            else:
                raise KeyError(f'Method "{method}" is invalid for model type "{self.model.__class__.__name__}"')

        return methods_list

    def to_session_data(self) -> ModelObjectGlueSessionData:
        return ModelObjectGlueSessionData(
            glue_type=self.glue_type,
            access=self.access,
            unique_name=self.unique_name,
            fields=self.generate_field_data(include_values=False).to_dict(),
            app_label=self.model_object._meta.app_label,
            model_name=self.model_object._meta.model_name,
            object_pk=self.model_object.pk,
            included_fields=self.included_fields,
            exclude_fields=self.excluded_fields,
            methods=self.generate_method_data(),
        )

    def update(self, updated_fields: dict[str, Any]):
        for key, value in updated_fields.items():
            if hasattr(self.model_object, key):
                setattr(self.model_object, key, value)

        self.model_object.save()

        self.fields = self.generate_field_data()



import json
from typing import Any

from django_glue.constants import UNIQUE_NAME_KEY, ACTION_KEY


class RequestBody:
    def __init__(self, django_request_body: bytes | str):
        self.data = json.loads(django_request_body.decode('utf-8'))
        self.unique_name = self.data[UNIQUE_NAME_KEY]
        self.action = self.data[ACTION_KEY]

    def __getitem__(self, key: str) -> Any:
        return self.data[key]

    def __setitem__(self, key: str, value: Any):
        self.data[key] = value



from django_glue.access.decorators import check_access
from django_glue.glue.model_object.actions import ModelObjectGlueAction
from django_glue.glue.model_object.response_data import MethodModelObjectGlueJsonData, ModelObjectGlueJsonData
from django_glue.glue.model_object.session_data import ModelObjectGlueSessionData
from django_glue.glue.model_object.tools import model_object_glue_from_session_data
from django_glue.glue.post_data import UpdatePostData, MethodPostData
from django_glue.handler.handlers import BaseRequestHandler
from django_glue.response.data import JsonResponseData
from django_glue.response.responses import generate_json_200_response_data


class GetModelObjectGlueHandler(BaseRequestHandler):
    action = ModelObjectGlueAction.GET
    _session_data_class = ModelObjectGlueSessionData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        model_object_glue = model_object_glue_from_session_data(self.session_data)
        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully retrieved model object!',
            data=ModelObjectGlueJsonData(model_object_glue.fields)
        )


class UpdateModelObjectGlueHandler(BaseRequestHandler):
    action = ModelObjectGlueAction.UPDATE
    _session_data_class = ModelObjectGlueSessionData
    _post_data_class = UpdatePostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        model_object_glue = model_object_glue_from_session_data(self.session_data)
        model_object_glue.update(self.post_data.fields)
        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully updated model object!',
            data=ModelObjectGlueJsonData(model_object_glue.fields)
        )


class DeleteModelObjectGlueHandler(BaseRequestHandler):
    action = ModelObjectGlueAction.DELETE
    _session_data_class = ModelObjectGlueSessionData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        model_object_glue = model_object_glue_from_session_data(self.session_data)
        model_object_glue.model_object.delete()
        return generate_json_200_response_data(
            message_title='Success',
            message_body='Successfully deleted model object!',
        )


class MethodModelObjectGlueHandler(BaseRequestHandler):
    action = ModelObjectGlueAction.METHOD
    _session_data_class = ModelObjectGlueSessionData
    _post_data_class = MethodPostData

    @check_access
    def process_response_data(self) -> JsonResponseData:
        model_object_glue = model_object_glue_from_session_data(self.session_data)
        method_return = model_object_glue.call_method(self.post_data.method, self.post_data.kwargs)
        return generate_json_200_response_data(
            'THE METHOD ACTION',
            'this is a response from an model object method action!',
            data=MethodModelObjectGlueJsonData(method_return)
        )



from django_glue.glue.function import handlers
from django_glue.glue.function.actions import FunctionGlueAction

FUNCTION_GLUE_HANDLER_MAP = {
    FunctionGlueAction.CALL: handlers.CallFunctionGlueHandler,
}



from django.conf import settings as django_settings

from django_glue import settings as django_glue_default_settings


class Settings:
    def __getattr__(self, name):
        if hasattr(django_settings, name):
            return getattr(django_settings, name)

        if hasattr(django_glue_default_settings, name):
            return getattr(django_glue_default_settings, name)

        raise f'No attribute {name} found in settings.'


settings = Settings()
